;; Generated package description from symbol-overlay.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "symbol-overlay" "4.3.0.20260423.145452" "Highlight symbols with keymap-enabled overlays" '((emacs "24.3") (seq "2.2")) :commit "253b957f5082603708b469d02ae8c31c58292823" :authors '(("wolray" . "wolray@foxmail.com")) :maintainer '("wolray" . "wolray@foxmail.com") :keywords '("faces" "matching") :url "https://github.com/wolray/symbol-overlay/")
